from blissoda.demo.id12 import id12_converter

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def id12_demo(expo=0.2, npoints=10):
    id12_converter.enable()
    try:
        setup_globals.loopscan(
            npoints,
            expo,
            setup_globals.diode1,
            setup_globals.diode2,
            setup_globals.mca1,
        )
    finally:
        id12_converter.disable()



"""For the BLISS demo session"""

import os

EWOKS_EVENTS_DIR = os.path.join(
    os.path.abspath(os.environ.get("DEMO_TMP_ROOT", ".")), "ewoks_events"
)

EWOKS_RESULTS_DIR = os.path.join(
    os.path.abspath(os.environ.get("DEMO_TMP_ROOT", ".")), "ewoks_results"
)



from ...exceptions import VersionError

try:
    # Bliss <=1.11
    # This needs to be tried before Bliss >=2
    from .ndarrayv0 import PersistentNdArray
except VersionError:
    try:
        # Bliss >=2
        from .ndarrayv1 import PersistentNdArray
    except VersionError as exc:
        _EXC = exc

        def _reraise(*_, **kw):
            raise _EXC

        class BlissDataVersionNotSupported:
            def __getattr__(self, *_):
                return _reraise

        PersistentNdArray = BlissDataVersionNotSupported



import logging
from typing import Iterator, Tuple, Dict

from ...exceptions import VersionError

try:
    from blissdata.scan import Scan
    from blissdata import redis_engine
    from blissdata.beacon.data import BeaconData
    from blissdata.redis_engine.search import get_next_scan
except ImportError as e:
    raise VersionError(str(e)) from e

logger = logging.getLogger(__name__)


def scan_iterator(session_name) -> Iterator[Tuple[str, int, Dict]]:
    logger.info(f"Started listening to Bliss session '{session_name}'")
    _ensure_redis()

    since = None
    while True:
        since, key = get_next_scan(since=since)
        scan = Scan.load(key)
        if scan.session != session_name:
            continue
        if scan.info.get("is-scan-sequence") or scan.info.get("group"):
            continue
        workflows = scan.info.get("workflows")
        if not workflows:
            continue
        filename = scan.info.get("filename")
        yield filename, scan.number, workflows


def _ensure_redis() -> None:
    url = BeaconData().get_redis_data_db()
    redis_engine.set_redis_url(url)



import os

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..id12.converter import Id12Hdf5ToAsciiConverter
from . import EWOKS_RESULTS_DIR


class DemoId12Hdf5ToSpecConverter(Id12Hdf5ToAsciiConverter):
    def __init__(self, **defaults) -> None:
        root_dir = os.path.join(EWOKS_RESULTS_DIR, "id12", "inhouse")
        defaults.setdefault(
            "external_proposal_outdir", os.path.join(root_dir, "EXTERNAL")
        )
        defaults.setdefault(
            "inhouse_proposal_outdir", os.path.join(root_dir, "INHOUSE2")
        )
        defaults.setdefault("test_proposal_outdir", os.path.join(root_dir, "NOBACKUP"))
        super().__init__(**defaults)

    def on_new_scan_metadata(self, scan) -> None:
        super().on_new_scan_metadata(scan)
        filename = self.output_filename(scan)
        dirname = os.path.dirname(filename)
        basename = os.path.basename(filename)
        scan_basename = f"scan{scan.scan_info['scan_nb']:03d}_{basename}"
        filename = os.path.join(dirname, scan_basename)
        print("Scan data will be saved in ASCII file", filename)


if setup_globals is None:
    id12_converter = None
else:
    id12_converter = DemoId12Hdf5ToSpecConverter()



from blissoda.xrpd.processor import XrpdProcessor
from blissoda.exafs.plotter import ExafsPlotter
from blissoda.streamline.scanner import StreamlineScanner

from blissoda.bm23.exafs_plotter import Bm23ExafsPlotter
from blissoda.bm02.xrpd_processor import Bm02XrpdProcessor
from blissoda.id11.xrpd_processor import Id11XrpdProcessor
from blissoda.id22.xrpd_processor import Id22XrpdProcessor
from blissoda.id31.xrpd_processor import Id31XrpdProcessor
from blissoda.id14.converter import Id14Hdf5ToSpecConverter
from blissoda.id12.converter import Id12Hdf5ToAsciiConverter
from blissoda.id31.streamline_scanner import Id31StreamlineScanner

from blissoda.demo.exafs import exafs_plotter
from blissoda.demo.xrpd import xrpd_processor
from blissoda.demo.id22 import stscan_processor
from blissoda.demo.id22 import id22_xrpd_processor
from blissoda.demo.id11 import id11_xrpd_processor
from blissoda.demo.id14 import id14_converter
from blissoda.demo.id12 import id12_converter
from blissoda.demo.streamline import streamline_scanner

from blissoda.demo.user_scripts.exafs import exafs_demo
from blissoda.demo.user_scripts.id22 import id22_stscan_demo
from blissoda.demo.user_scripts.id22 import id22_xrpd_demo
from blissoda.demo.user_scripts.streamline import streamline_demo
from blissoda.demo.user_scripts.xrpd import xrpd_demo
from blissoda.demo.user_scripts.id11 import id11_pdf_demo  # noqa F401
from blissoda.demo.user_scripts.id14 import id14_demo
from blissoda.demo.user_scripts.id12 import id12_demo

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def all_print():
    _print_objects(exafs_plotter)
    _print_objects(xrpd_processor)
    _print_objects(id11_xrpd_processor)
    _print_objects(id22_xrpd_processor)
    _print_objects(id14_converter)
    _print_objects(id12_converter)
    _print_objects(stscan_processor)
    _print_objects(streamline_scanner)

    _print_objects(XrpdProcessor())
    _print_objects(ExafsPlotter())
    _print_objects(StreamlineScanner())
    _print_objects(Bm23ExafsPlotter())
    _print_objects(Bm02XrpdProcessor())
    _print_objects(Id11XrpdProcessor())
    _print_objects(Id22XrpdProcessor())
    _print_objects(Id31XrpdProcessor())
    _print_objects(Id14Hdf5ToSpecConverter())
    _print_objects(Id12Hdf5ToAsciiConverter())

    _print_objects(Id31StreamlineScanner())

    print()
    print("SUCCESS: all objects can be printed")


def all_demo():
    print()
    print("===================")
    setup_globals.newcollection("id22_stscan_collection")
    id22_stscan_demo()

    print()
    print("===================")
    setup_globals.newcollection("id22_xrpd_collection")
    id22_xrpd_demo()

    print()
    print("===================")
    setup_globals.newcollection("streamline_collection")
    streamline_demo()

    print()
    print("===================")
    setup_globals.newcollection("id11_collection")
    print("TODO: get license for diffpy")
    # id11_pdf_demo()

    print()
    print("===================")
    setup_globals.newcollection("xrpd_collection")
    xrpd_demo()

    # TODO: ewoksxrpd needs to be imported before xraylarch?
    #       Cannot load backend 'Qt5Agg' which requires the 'qt' interactive framework
    print()
    print("===================")
    setup_globals.newcollection("exafs_collection")
    exafs_demo()

    print()
    print("===================")
    setup_globals.newcollection("id14_collection")
    id14_demo()

    print()
    print("===================")
    setup_globals.newcollection("id12_collection")
    id12_demo()

    print()
    print("===================")
    setup_globals.newcollection("done_collection")

    print()
    print("SUCCESS: all demos can be executed")


def _print_objects(obj):
    print()
    print("===================")
    print(obj._parameters.name)
    print(obj.__info__())



"""Automatic pyfai integration for every scan with saving and plotting with pdf extraction"""

import os
import tempfile
import configparser
from glob import glob
from typing import Optional, Dict, List

from ..xrpd.processor import XrpdProcessor
from ..persistent.parameters import ParameterInfo
from ..utils.pyfai import read_config

from ..resources.id11 import RESOURCE_ROOT


class Id11XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config_directory", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("pdf_enable", category="PDF"),
        ParameterInfo("pdf_options", category="PDF"),
        ParameterInfo("pdf_config_file", category="PDF"),
    ],
):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("pyfai_config_directory", tempfile.gettempdir())
        defaults.setdefault("pdf_config_file", tempfile.gettempdir())
        super().__init__(**defaults)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        if not self.lima_names:
            return categories
        categories["PyFai"]["integration_options"] = "... (see below)"
        lima_name = self.lima_names[0]
        categories["PyFai integration"] = {
            "1. JSON file": self._detector_config_filename(lima_name, ".json"),
            "2. PONI file": self._detector_config_filename(lima_name, ".poni"),
            "3. User": self.integration_options,
            "Merged": self.get_integration_options(lima_name),
        }
        categories["PDF"] = {
            "PDF Enable": self.pdf_enable,
            "PDF Config File Path": self.pdf_config_file,
            "PDF Config From User": self.pdf_options,
            "PDFGetX Config Parameters Merged": self.get_pdfgetx_options(lima_name),
        }
        return categories

    def _set_parameter(self, name, value):
        super()._set_parameter(name, value)
        if name == "pdf_enable":
            self._update_pdfgetx_status()

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return None

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        options = self._default_azint_options(lima_name)
        options.update(self._default_calib_options(lima_name))
        integration_options = self.integration_options
        if integration_options:
            options.update(integration_options.to_dict())
        return options

    def _default_azint_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".json")
        return read_config(filename)

    def _default_calib_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".poni")
        return read_config(filename)

    def _detector_config_filename(self, lima_name: str, ext: str) -> Optional[str]:
        pyfai_config_directory = self.pyfai_config_directory
        if not pyfai_config_directory:
            pyfai_config_directory = tempfile.gettempdir()
        pattern = os.path.join(pyfai_config_directory, lima_name, f"*{ext}")
        files = sorted(glob(pattern))
        if not files:
            return
        return files[-1]

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        if self.pdf_enable:
            inputs += self._get_pdfgetx_config_inputs(lima_name)
            inputs += self._get_pdfgetx_save_inputs(scan, lima_name)
        return inputs

    def _update_pdfgetx_status(self):
        if self.pdf_enable:
            self.workflow_with_saving = os.path.join(
                RESOURCE_ROOT, "integrate_scan_with_saving_pdf.json"
            )
        else:
            self.workflow_with_saving = os.path.join(
                RESOURCE_ROOT, "integrate_scan_with_saving.json"
            )
        return self.pdf_enable

    def get_pdfgetx_options(self, lima_name: str) -> Optional[dict]:
        config = configparser.ConfigParser()
        config.read(self.pdf_config_file)
        options = dict(config["DEFAULT"])
        for key in options.keys():  # Patch config parser str -> float
            if options[key].replace(".", "").isnumeric():
                options[key] = float(options[key])
        pdf_options = self.pdf_options
        if pdf_options:
            options.update(pdf_options.to_dict())
        return options

    def _get_pdfgetx_config_inputs(self, lima_name: str) -> List[dict]:
        inputs = [
            {
                "task_identifier": "PdfGetXConfig",
                "name": "pdfgetx_options_dict",
                "value": self.get_pdfgetx_options(lima_name),
            }
        ]
        return inputs

    def _get_pdfgetx_save_inputs(self, scan, lima_name: str) -> List[dict]:
        scan_nb = scan.scan_info.get("scan_nb")
        filename = self.save_nexus_filename(scan, lima_name=lima_name)
        save_path = filename.replace("%s.h5" % lima_name, "%s.h5" % scan_nb)
        inputs = [
            {
                "task_identifier": "PdfGetXSaveAscii",
                "name": "filename",
                "value": save_path,
            }
        ]
        inputs += [
            {
                "task_identifier": "PdfGetXSaveNexus",
                "name": "output_filename",
                "value": self.save_nexus_filename(scan),
            },
            {
                "task_identifier": "PdfGetXSaveNexus",
                "name": "scan",
                "value": scan_nb,
            },
            {
                "task_identifier": "PdfGetXSaveNexus",
                "name": "detector_name",
                "value": lima_name,
            },
        ]
        return inputs



from blissoda.demo.id14 import id14_converter

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def id14_demo(expo=0.2, npoints=10):
    id14_converter.enable()
    try:
        setup_globals.loopscan(npoints, expo, setup_globals.mca1)
    finally:
        id14_converter.disable()



import os
import json
from pprint import pprint
from typing import List

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
try:
    from bliss import current_session
except ImportError:
    current_session = None

from .calib import DEFAULT_CALIB
from ..id22.stscan_processor import StScanProcessor
from ..id22.xrpd_processor import Id22XrpdProcessor
from ..utils import directories


class DemoStScanProcessor(StScanProcessor):
    def __init__(self, **defaults) -> None:
        root_dir = directories.get_processed_dir(current_session.scan_saving.filename)
        root_dir = os.path.join(root_dir, "accessible", "id22", "config")

        defaults.setdefault(
            "_convert_workflow",
            os.path.join(root_dir, "convert.json"),
        )
        defaults.setdefault(
            "_rebinsum_workflow",
            os.path.join(root_dir, "rebinsum.json"),
        )
        defaults.setdefault(
            "_extract_workflow",
            os.path.join(root_dir, "extract.json"),
        )
        super().__init__(**defaults)

    def _submit_job(self, workflow, inputs, convert_destination, **kw):
        print("\nSubmit workfow")
        print(workflow)
        print("Inputs:")
        pprint(inputs)
        print("Save for provenance:")
        pprint(convert_destination)
        print("Options:")
        pprint(kw)

    def _get_workflows_dir(self, dataset_filename: str) -> str:
        return os.path.join(directories.get_workflows_dir(dataset_filename), "id22")


class DemoId22XrpdProcessor(Id22XrpdProcessor):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("lima_names", ["difflab6"])
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "radial_range_min": 1,
                "unit": "q_nm^-1",
            },
        )
        super().__init__(**defaults)
        self._ensure_config_filename()

    def get_integrate_1d_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_integrate_1d_inputs(scan, lima_name)
        is_even = not bool(
            setup_globals.difflab6.image.width % 2
        )  # lima-camera-simulator<1.9.10 does not support odd image widths
        inputs.append(
            {"task_identifier": "Integrate1D", "name": "demo", "value": is_even}
        )
        return inputs

    def _ensure_config_filename(self):
        if self.pyfai_config:
            return
        root_dir = self._get_accessible_config_dir(current_session.scan_saving.filename)
        cfgfile = os.path.join(root_dir, "pyfaicalib.json")
        os.makedirs(os.path.dirname(cfgfile), exist_ok=True)
        poni = DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.pyfai_config = cfgfile

    def _get_accessible_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "accessible", "id22")

    def _get_accessible_workflows_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_accessible_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "config")


if setup_globals is None:
    stscan_processor = None
    id22_xrpd_processor = None
else:
    stscan_processor = DemoStScanProcessor()
    id22_xrpd_processor = DemoId22XrpdProcessor()



import os
import json
from typing import Tuple, Any, Generator

import numpy

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
try:
    from bliss import current_session
except ImportError:
    current_session = None

from ..streamline.scanner import StreamlineScanner
from ..utils import directories


def energy_wavelength(x):
    """keV to m and vice versa"""
    return 12.398419843320026 * 1e-10 / x


_DEFAULT_CALIB = {
    "dist": 5e-2,  # 5 cm
    "poni1": 10e-2,  # 10 cm
    "poni2": 10e-2,  # 10 cm
    "rot1": numpy.radians(10),  # 10 deg
    "rot2": 0,  # 0 deg
    "rot3": 0,  # 0 deg
    "wavelength": energy_wavelength(12),  # 12 keV
    "detector": "Pilatus1M",
}


class DemoStreamlineScanner(StreamlineScanner):
    def __init__(self, **defaults):
        defaults.setdefault("detector_name", "difflab6")
        defaults.setdefault("calibrant", "LaB6")
        defaults.setdefault("sample_changer_name", "streamline_sc")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "integrator_name": "sigma_clip_ng",
                "extra_options": {"max_iter": 3, "thres": 0},
                "error_model": "azimuthal",  # hybrid gives weird results
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        super().__init__(**defaults)

    def measure_sample(self, *args, has_qrcode: bool = True, **kwargs):
        return setup_globals.sct(*args, setup_globals.difflab6, **kwargs)

    def calib(self, *args, sample_index=0, **kwargs):
        return super().calib(*args, sample_index=sample_index, **kwargs)

    def init_workflow(self, with_autocalibration: bool = True):
        self._ensure_pyfai_config()
        return super().init_workflow(with_autocalibration=with_autocalibration)

    def _ensure_pyfai_config(self):
        """Set the pyfai_config variable and ensure the file exists."""
        if not self.pyfai_config:
            root_dir = self._get_accessible_config_dir(
                current_session.scan_saving.filename
            )
            self.pyfai_config = os.path.join(root_dir, "pyfaicalib.json")
        if os.path.exists(self.pyfai_config):
            return
        os.makedirs(os.path.dirname(self.pyfai_config), exist_ok=True)
        poni = _DEFAULT_CALIB
        with open(self.pyfai_config, "w") as f:
            json.dump(poni, f)

    def _get_accessible_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "accessible", "streamline")

    def _get_workflows_dirname(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_accessible_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "config")

    def run(self, *args, **kwargs):
        self._ensure_pyfai_config()  # in case the file was deleted
        if not os.path.exists(self.workflow):
            raise RuntimeError(
                "the workflow file no longer exists, execute 'init_workflow' again"
            )
        super().run(*args, **kwargs)

    @property
    def sample_changer(self):
        return streamline_sc

    def _job_arguments(self, *args, **kw):
        args, kwargs = super()._job_arguments(*args, **kw)
        is_even = not bool(
            setup_globals.difflab6.image.width % 2
        )  # lima-camera-simulator<1.9.10 does not support odd image widths
        kwargs["inputs"].append(
            {"task_identifier": "Integrate1D", "name": "demo", "value": is_even}
        )
        return args, kwargs

    def _get_workflow_upload_parameters(self, *args) -> None:
        return None


class MockQrReader:
    QRCODE_NOT_READABLE = "QRCODE_NOT_READABLE"


class MockSampleChanger:
    def __init__(self, number_of_samples: int = 16) -> None:
        translation = MockSampleTranslation()

        self._vibration_speed = 0
        self._translation = translation
        self._number_of_samples = number_of_samples
        self._allowed_sample_indices = tuple(range(number_of_samples))
        self._qrreader = MockQrReader()

        # should be in Redis but good enough for a demo
        self._nholders_in_tray = 2
        self._loaded = False
        self._holder_counter = 0

        start = 0
        stop = number_of_samples - 1
        intervals = number_of_samples - 1
        self._ascan_arguments = translation, start, stop, intervals

    @property
    def qrreader(self) -> MockQrReader:
        return self._qrreader

    def ascan_arguments(self) -> Tuple[Any, float, float, int]:
        return self._ascan_arguments

    def fill_tray(self, n=2) -> None:
        self._nholders_in_tray = n

    def qr_read(self, **_) -> str:
        if not self._loaded:
            return self.qrreader.QRCODE_NOT_READABLE
        position = self._translation.position
        sample_index = int(position)
        on_sample = abs(position - sample_index) < 0.01
        if not on_sample or sample_index not in self._allowed_sample_indices:
            return self.qrreader.QRCODE_NOT_READABLE
        return f"holder{self._holder_counter}_sample{sample_index:02d}_lab6"

    def select_sample(self, sample_index: int, **qrread_options) -> str:
        for qr_response in self.iterate_samples([sample_index], **qrread_options):
            return qr_response

    def select_sample_without_qr(self, sample_index) -> str:
        return self.select_sample(sample_index)

    def iterate_samples(
        self, sample_indices=None, **qrread_options
    ) -> Generator[str, None, None]:
        if not sample_indices:
            sample_indices = range(self._number_of_samples)
        for i in sample_indices:
            self.translation.position = i
            yield self.qr_read(**qrread_options)

    def iterate_samples_without_qr(
        self, sample_indices=None
    ) -> Generator[str, None, None]:
        yield from self.iterate_samples(sample_indices=sample_indices)

    def eject_old_baguette(self) -> None:
        if not self._loaded:
            print("Streamline sample changer: nothing loaded")
            return
        self._loaded = False
        self.translation.position = 20

    def load_baguette_with_homing(self) -> None:
        if self._loaded:
            print("Streamline sample changer: already loaded")
            return
        if self._nholders_in_tray == 0:
            print("Streamline sample changer: no more holders in tray")
            return
        self._nholders_in_tray = self._nholders_in_tray - 1
        self._loaded = True
        self._holder_counter += 1
        self.translation.position = 0

    def has_remaining_baguettes(self) -> bool:
        return bool(self._nholders_in_tray)

    @property
    def number_of_remaining_baguettes(self) -> int:
        return self._nholders_in_tray

    @property
    def translation(self):
        return self._translation

    @property
    def vibration_speed(self):
        return self._vibration_speed

    @vibration_speed.setter
    def vibration_speed(self, speed):
        print("SETTING VIBRATION SPEED TO", speed)
        if speed < 0 or speed > 100:
            raise RuntimeError(
                "Speed for the fluidization system out of range (0-100, as in %)"
            )
        self._vibration_speed = speed

    def tune_qrreader(self, force=False) -> str:
        return self.qr_read()

    def tune_qrreader_for_baguette(self) -> None:
        pass


class MockSampleTranslation:
    def __init__(self) -> None:
        self._position = 0

    def on(self):
        pass

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        print("Move streamline_translation from", self._position, "to", value)
        self._position = value


if setup_globals is None:
    streamline_scanner = None
    streamline_sc = None
    streamline_translation = None
else:
    streamline_scanner = DemoStreamlineScanner()
    streamline_sc = MockSampleChanger()
    streamline_translation = MockSampleTranslation()



import os
import json
from typing import List

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
try:
    from bliss import current_session
except ImportError:
    current_session = None

from .calib import DEFAULT_CALIB
from ..resources.demo import RESOURCE_ROOT
from ..id11.xrpd_processor import Id11XrpdProcessor
from ..utils import directories


class DemoId11XrpdProcessor(Id11XrpdProcessor):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("lima_names", ["difflab6"])
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "radial_range_min": 1,
                "unit": "q_A^-1",
            },
        )
        super().__init__(**defaults)
        self._ensure_pyfai_config_filename()
        self._ensure_pdf_config_filename()

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        inputs = super().get_integrate_inputs(scan, lima_name, task_identifier)
        is_even = not bool(
            setup_globals.difflab6.image.width % 2
        )  # lima-camera-simulator<1.9.10 does not support odd image widths
        inputs.append(
            {"task_identifier": task_identifier, "name": "demo", "value": is_even}
        )
        return inputs

    def _ensure_pyfai_config_filename(self):
        root_dir = self._get_accessible_config_dir(current_session.scan_saving.filename)
        cfgfile = os.path.join(root_dir, "difflab6", "pyfaicalib.json")
        if not os.path.exists(cfgfile):
            poni = DEFAULT_CALIB
            os.makedirs(os.path.dirname(cfgfile), exist_ok=True)
            with open(cfgfile, "w") as f:
                json.dump(poni, f)
        self.pyfai_config_directory = root_dir

    def _ensure_pdf_config_filename(self):
        src_pdf_config_file = os.path.join(RESOURCE_ROOT, "pdf_config.cfg")
        root_dir = self._get_accessible_config_dir(current_session.scan_saving.filename)
        dest_pdf_config_file = os.path.join(root_dir, "pdf_config.cfg")
        if not os.path.exists(dest_pdf_config_file):
            with open(src_pdf_config_file, "r") as src:
                os.makedirs(os.path.dirname(dest_pdf_config_file), exist_ok=True)
                with open(dest_pdf_config_file, "w") as dest:
                    src_contents = src.read()
                    dest.write(src_contents)
                    dest.write(
                        f'backgroundfile = {os.path.join(RESOURCE_ROOT, "background.xy")}'
                    )
        self.pdf_config_file = dest_pdf_config_file

    def _get_accessible_dir(self, dataset_filename: str) -> str:
        root_dir = directories.get_processed_dir(dataset_filename)
        return os.path.join(root_dir, "accessible", "id11")

    def _get_accessible_workflows_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "workflows")

    def _get_accessible_config_dir(self, dataset_filename: str) -> str:
        root_dir = self._get_accessible_dir(dataset_filename)
        return os.path.join(root_dir, "config")


if setup_globals is None:
    id11_xrpd_processor = None
else:
    id11_xrpd_processor = DemoId11XrpdProcessor()



import os
import mock
import pytest
from copy import deepcopy
from contextlib import ExitStack
from collections import namedtuple
from collections.abc import MutableMapping

from typing import Any, Iterator, Mapping
from ..persistent.parameters import WithPersistentParameters


def has_redis_server() -> bool:
    # TODO: instead of pytest-redis which needs a real Redis server, use fakeredis?
    with os.popen("redis-server --version") as output:
        return bool(output.read())


MockedBlissInfo = namedtuple("MockedBlissInfo", "beacon_connection")

if has_redis_server():

    @pytest.fixture(scope="session")
    def mock_bliss_session(redis_proc):
        # Ensure we do not connect to a real Beacon host
        os.environ["BEACON_HOST"] = "nonexisting:25000"

        # Skip test when Beacon dependencies are not available
        # TODO: mock to the extend we no longer need bliss as a dependency
        try:
            try:
                from bliss.redis.manager import RedisAddress
            except ImportError:
                from blissdata.redis import RedisAddress
            from bliss.config.conductor.connection import Connection
            from bliss.config.conductor.client import get_default_connection
        except ImportError:
            raise
            pytest.skip("requires bliss")

        redis_url = f"{redis_proc.host}:{redis_proc.port}"

        def _get_redis_url(self):
            """Return the pytest-redis server URL"""
            nonlocal redis_url
            return RedisAddress.factory(redis_url)

        with ExitStack() as stack:
            # The Beacon connection should connect to the pytest-redis server
            ctx = mock.patch.object(
                Connection, "get_redis_connection_address", _get_redis_url
            )
            stack.enter_context(ctx)
            ctx = mock.patch.object(
                Connection, "get_redis_data_server_connection_address", _get_redis_url
            )
            stack.enter_context(ctx)

            yield MockedBlissInfo(get_default_connection())

else:

    @pytest.fixture(scope="session")
    def mock_bliss_session():
        # redis_proc raises an exception when redis-server is not available
        pytest.skip("requires redis-server")


@pytest.fixture
def mock_bliss(mock_bliss_session):
    yield
    proxy = mock_bliss_session.beacon_connection.get_redis_proxy()
    proxy.flushall()


@pytest.fixture
def mock_persistent():
    # TODO: remove the need for this when we eliminated the need to a Bliss dependency
    remote_dict = MockHashObjSetting()

    def init(self, **defaults):
        for name, value in defaults.items():
            remote_dict.setdefault(name, value)
        self._parameters = remote_dict

    with mock.patch.object(WithPersistentParameters, "__init__", init):
        yield remote_dict


class MockHashObjSetting(MutableMapping):
    def __init__(self) -> None:
        self._adict = dict()
        super().__init__()

    def __repr__(self) -> str:
        return repr(self._adict)

    def get_all(self) -> dict:
        return deepcopy(self._adict)

    def __getitem__(self, key: str) -> Any:
        return self._adict[key]

    def __setitem__(self, key: str, value: Any) -> None:
        if isinstance(value, Mapping):
            value = deepcopy(value)
        self._adict[key] = value

    def __delitem__(self, key: str) -> None:
        del self._adict[key]

    def __iter__(self) -> Iterator[Any]:
        yield from self._adict

    def __len__(self) -> int:
        return len(self._adict)



"""User API for HDF5 conversion on the Bliss repl"""

import os
from typing import List, Dict

from ewoksjob.client import submit

try:
    from bliss import current_session
except ImportError:
    current_session = None

from ..utils import trigger
from ..utils import directories
from ..persistent.parameters import ParameterInfo
from ..persistent.parameters import WithPersistentParameters


class Id14Hdf5ToSpecConverter(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("_enabled"),
        ParameterInfo("workflow", category="workflows"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("worker", category="workflows"),
    ],
):
    def __init__(self, **defaults) -> None:
        if current_session is None:
            raise ImportError("bliss")
        defaults.setdefault("_enabled", False)
        defaults.setdefault("worker", "celery")
        defaults.setdefault(
            "workflow", "/data/id14/inhouse/ewoks/resources/workflows/convert.json"
        )
        super().__init__(**defaults)
        self._sync_scan_metadata()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        categories["status"] = {"Enabled": self._enabled}
        return categories

    def enable(self, *detectors) -> None:
        self._enabled = True
        self._sync_scan_metadata()

    def disable(self) -> None:
        self._enabled = False
        self._sync_scan_metadata()

    def _sync_scan_metadata(self) -> None:
        if self._enabled:
            workflows_category = trigger.register_workflow_category()
            workflows_category.set("processing", self.on_new_scan_metadata)
        else:
            trigger.unregister_workflow_category()

    def on_new_scan_metadata(self, scan) -> None:
        if not self.scan_requires_processing(scan):
            return
        kwargs = self.get_submit_arguments(scan)
        _ = submit(args=(self.workflow,), kwargs=kwargs, queue=self.worker)

    def scan_requires_processing(self, scan) -> bool:
        # TODO: select scan that needs processing
        return True

    def get_submit_arguments(self, scan) -> dict:
        return {
            "inputs": self.get_inputs(scan),
            "outputs": [{"all": False}],
        }

    def get_inputs(self, scan) -> List[dict]:
        task_identifier = "Hdf5ToSpec"

        filename = self.get_filename(scan)
        output_filename = self.workflow_destination(scan)
        scan_nb = scan.scan_info.get("scan_nb")

        # Scan metadata published in id14.McaAcq.McaAcq.save
        calibration = scan.scan_info["instrument"]["calibration"]
        mca_calibration = calibration["a"], calibration["b"], 0

        return [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": output_filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "scan_numbers",
                "value": [scan_nb],
            },
            {
                "task_identifier": task_identifier,
                "name": "mca_calibration",
                "value": mca_calibration,
            },
        ]

    def get_filename(self, scan) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def workflow_destination(self, scan) -> str:
        filename = self.get_filename(scan)
        root = directories.get_processed_dir(filename)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}.mca"
        return os.path.join(root, basename)

    def enable_slurm(self):
        self.worker = "slurm"

    def disable_slurm(self):
        self.worker = "celery"



try:
    from bliss.scanning.scan_meta import get_user_scan_meta
    from bliss.scanning.scan_meta import META_TIMING
except ImportError:

    def get_user_scan_meta():
        raise ImportError("bliss is required")


def register_workflow_category(category: str = "workflows", timing: str = None):
    scan_meta_obj = get_user_scan_meta()
    if category not in scan_meta_obj.categories_names():
        scan_meta_obj.add_categories({category})
    if timing:
        timing = META_TIMING[timing.upper()]
    else:
        timing = META_TIMING.PREPARED
    scan_meta_obj.workflows.timing = timing
    scan_meta_obj.workflows.set("@NX_class", {"@NX_class": "NXcollection"})
    return scan_meta_obj.workflows


def unregister_workflow_category(category: str = "workflows") -> None:
    scan_meta_obj = get_user_scan_meta()
    try:
        # Before Bliss 1.11
        keys = list(scan_meta_obj._metadata.keys())
        for key in keys:
            if key.name.lower() == category:
                scan_meta_obj._metadata.pop(key)
    except Exception:
        pass
    # Since Bliss 1.11
    scan_meta_obj.remove_categories({category})
    try:
        _ = list(scan_meta_obj._metadata_keys())
    except AttributeError:
        pass



__version__ = "0.4.1"

try:
    from gevent.monkey import is_anything_patched
    from gevent.monkey import is_module_patched
except ImportError:
    pass
else:
    if is_anything_patched() and not is_module_patched("threading"):
        try:
            # comes with ewoksjob (celery):
            from kombu.utils import compat
        except ImportError:
            pass
        else:
            # Make Celery use `celery.backends.asynchronous.Drainer`
            # instead of `celery.backends.asynchronous.geventDrainer`.
            # The later causes CTRL-C to not be raised and other things
            # like Bliss scans to hang when calling `AsyncResult.get`.

            compat._environment = "default"

            # The real solution is to patch threads.



import pickle
from typing import Union, List, Dict

import numpy

from ...exceptions import VersionError

try:
    from bliss.config.conductor.client import get_redis_proxy
except ImportError:
    try:
        from blissdata.client import get_redis_proxy
    except ImportError as e:
        raise VersionError(str(e)) from e


class PersistentNdArray:
    def __init__(self, redis_key: str) -> None:
        self._connection = get_redis_proxy(db=0)
        self._name = redis_key

    def append(self, value: numpy.ndarray) -> None:
        self._connection.xadd(self._name, _encode(value[numpy.newaxis, ...]))

    def extend(self, value: numpy.ndarray) -> None:
        self._connection.xadd(self._name, _encode(value))

    def __getitem__(self, idx) -> Union[numpy.ndarray, List[numpy.ndarray]]:
        if idx == 0:
            adict = self._connection.xrange(self._name, count=1)[0][1]
            return _decode(adict)[0]
        if idx == -1:
            adict = self._connection.xrevrange(self._name, count=1)[0][1]
            return _decode(adict)[-1]
        arr = [_decode(adict) for _, adict in self._connection.xrange(self._name)]
        if arr:
            arr = numpy.vstack(arr)
        else:
            arr = numpy.array([])
        if isinstance(idx, tuple) and not idx:
            return arr
        return arr[idx]

    def remove(self) -> None:
        self._connection.delete(self._name)


def _encode(data: numpy.ndarray) -> Dict[bytes, bytes]:
    return {b"data": pickle.dumps(data)}


def _decode(data: Dict[bytes, bytes]) -> numpy.ndarray:
    return pickle.loads(data[b"data"])



from typing import Any, Union

import json
import pickle

try:
    from blissdata.settings import OrderedHashObjSetting
except ImportError:
    try:
        from bliss.config.settings import OrderedHashObjSetting
    except ImportError:
        OrderedHashObjSetting = None


class PersistentOrderedDict(OrderedHashObjSetting):
    def __init__(self, redis_key: str, serializer: str = "json"):
        dumps, loads = _SERIALIZERS[serializer]
        super().__init__(
            redis_key, read_type_conversion=loads, write_type_conversion=dumps
        )


def _json_loads(serialized_value: Union[bytes, Any]) -> Any:
    # Called twice on the same data for bliss<2
    if isinstance(serialized_value, bytes):
        try:
            return json.loads(serialized_value)
        except Exception:
            pass
    return serialized_value


def _pickle_loads(serialized_value: Union[bytes, Any]) -> Any:
    # Called twice on the same data for bliss<2
    if isinstance(serialized_value, bytes):
        try:
            return pickle.loads(serialized_value)
        except Exception:
            pass
    return serialized_value


def _json_dumps(python_value: Any) -> str:
    return json.dumps(python_value)


def _pickle_dumps(python_value: Any) -> bytes:
    return pickle.dumps(python_value)


_SERIALIZERS = {
    "json": (_json_dumps, _json_loads),
    "pickle": (_pickle_dumps, _pickle_loads),
}
